/*
 * routewatch
 *
 * Monitor DVMRP route updates on a multicast router
 * (or on a machine on the same subnet)
 *
 * If you run it on a router you can monitor all of that router's tunnels
 * If you run it on a non-router you can only monitor that subnet's routing.
 *
 *
 * Copyright (c) Xerox Corporation 1997, 1998. All rights reserved.
 *
 * License is granted to copy, to use, and to make and to use derivative
 * works, provided that Xerox is acknowledged in all documentation
 * pertaining to any such copy or derivative work. Xerox grants no other
 * licenses expressed or implied. The Xerox trade name should not be used
 * in any advertising without its written permission.
 *
 * XEROX CORPORATION MAKES NO REPRESENTATIONS CONCERNING EITHER THE
 * MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS SOFTWARE
 * FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" without
 * express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */
#include <stdio.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/igmp.h>
#include <arpa/inet.h>

char *formats[] = { "%d", "%d.%d", "%d.%d.%d", "%d.%d.%d.%d" };
	
void
usage()
{
	fprintf(stderr, "Usage: routewatch [-r p.p.p.p] [-i i.i.i.i] [-cet] [r.r.r.r/m]\n");
	exit(1);
}

/*
 * Set the socket buffer.  If we can't set it as large as we
 * want, search around to try to find the highest acceptable
 * value.  The highest acceptable value being smaller than
 * minsize is a fatal error.
 */
void k_set_rcvbuf(s, bufsize, minsize)
    int s;
    int bufsize;
    int minsize;
{
    int delta = bufsize / 2;
    int iter = 0;

    if (setsockopt(s, SOL_SOCKET, SO_RCVBUF,
		(char *)&bufsize, sizeof(bufsize)) < 0) {
	bufsize -= delta;
	while (1) {
	    iter++;
	    if (delta > 1)
		delta /= 2;

	    if (setsockopt(s, SOL_SOCKET, SO_RCVBUF,
			(char *)&bufsize, sizeof(bufsize)) < 0) {
		    bufsize -= delta;
	    } else {
		    if (delta < 1024)
			break;
		    bufsize += delta;
	    }
	}
	if (bufsize < minsize) {
	    fprintf(stderr,"OS-allowed buffer size %u < app min %u\n",
		bufsize, minsize);
	    exit(1);
	}
    }
    printf("Got %d byte buffer size in %d iterations\n",
	    bufsize, iter);
}

int
main(argc, argv)
	int argc;
	char **argv;
{
	int s;
	unsigned char buf[2048];
	struct ip *ip = (struct ip *)buf;
	struct igmp *igmp;
	struct sockaddr_in recv;
	int recvlen = sizeof(recv);
	unsigned char *p, *e;
	int i, pktlen, couldsave;
	unsigned int addr, mask, oldmask;
	unsigned int wantaddr, wantmask;
	int metric, masklen;
	struct timeval t;
	int exactmatch = 0;
	char logbuf[100], *lbp;
	int u;
	struct sockaddr_in sandbox;
	struct ip_mreq mreq;
	int pktno = 0;
	int rptno;
	struct tm *tt;
	int cflag = 0;
	int tflag = 0;
	int rflag = 0;
	unsigned int nbraddr = INADDR_ANY;
	unsigned int interface = INADDR_ANY;
	int ch;
	extern int optind;
	extern char *optarg;

	/*
	 * Default to accepting all routes
	 */
	wantaddr = 0x00000000;
	wantmask = 0x00000000;
	/*
	 * Parse arguments.
	 */
	while ((ch = getopt(argc, argv, "cei:r:t")) != -1)
		switch (ch) {
		/*
		 * -c : print possible compression for each route update
		 */
		case 'c':
			cflag++;
			break;

		/*
		 * -e : only match exactly the route specified
		 */
		case 'e':
			exactmatch = 1;
			break;

		case 'i':
			if ((interface = inet_addr(optarg)) == -1) {
				fprintf(stderr, "can't parse %s as ipaddr\n",
					optarg);
				usage();
				/*NOTREACHED*/
			}
			break;

		/*
		 * -r : only match routes from this neighbor
		 */
		case 'r':
			if ((nbraddr = inet_addr(optarg)) == -1) {
				fprintf(stderr, "can't parse %s as ipaddr\n",
					optarg);
				usage();
				/*NOTREACHED*/
			}
			rflag++;
			break;

		/*
		 * -t : print unix timestamps
		 */
		case 't':
			tflag++;
			break;

		default:
			usage();
			/*NOTREACHED*/
		}
	optind--;
	argc -= optind;
	argv += optind;
	
	/*
	 * Any remaining argument is the route we want to monitor.
	 */
	if (argc > 1) {
		int a[4];
		int n;

		if ((n = sscanf(argv[1], "%d.%d.%d.%d",
				&a[0], &a[1], &a[2], &a[3])) < 1)
			usage();
			/*NOTREACHED*/
		for (i = 0; i < n; i++)
			wantaddr = wantaddr << 8 | a[i];
		for (; i < 4; i++)
			wantaddr = wantaddr << 8;
		n = -1;
		if (sscanf(argv[1], "%*[^/]/%d", &n) != 1 || n < 0 || n > 32)
			usage();
			/*NOTREACHED*/
		wantmask = 0xffffffff;
		for (i = 32; i > n; i--)
			wantmask = wantmask << 1;

		wantmask = wantmask & 0xffffffff;
	}
	wantaddr &= wantmask;
	printf("wantaddr = %x, wantmask = %x\n", wantaddr, wantmask);
#ifdef SYSLOG
	/*
	 * Open and configure a socket to send syslog packets on.
	 */
	if ((u = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("udp socket");
		exit(1);
	}
	sandbox.sin_family = AF_INET;
	sandbox.sin_port = htons(514);	/* logger */
	sandbox.sin_addr.s_addr = htonl(0xc0d83601);	/* sandbox.xerox.com */
	if (connect(u, &sandbox, sizeof(sandbox)) < 0) {
		perror("connect");
		exit(1);
	}
	sprintf(logbuf, "<%d>", LOG_LOCAL5 | LOG_INFO);
	lbp = logbuf + strlen(logbuf);
#else
	/*
	 * Just print to stdout.
	 */
	u = 1;	/* stdout */
#endif
	/*
	 * Open a raw IGMP socket.
	 */
	if ((s=socket(AF_INET, SOCK_RAW, IPPROTO_IGMP)) < 0) {
		perror("igmp socket (r u root?)");
		exit(1);
	}
#ifdef SAFE
	setuid(-2);	/* nobody */
#endif
	/*
	 * Join the multicast group on the raw socket.
	 */
	mreq.imr_multiaddr.s_addr = htonl(0xe0000004);	/* DVMRP.MCAST.NET */
	mreq.imr_interface.s_addr = htonl(interface);
	if (setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void *)&mreq, sizeof(mreq)) < 0) {
	    	perror("IP_ADD_MEMBERSHIP");
		exit(1);
	}
	/*
	 * Set the socket buffer to the largest N  48k <= N <= 256k
	 * that the OS will allow.
	 */
	k_set_rcvbuf(s, 256*1024, 48*1024);
	igmp = (struct igmp *)(ip + 1);
	while (1) {
		/*
		 * Receive a packet.
		 */
		pktlen = recvfrom(s, buf, sizeof(buf), 0,
				    (struct sockaddr *)&recv, &recvlen);
		/*
		 * Make sure it's from the right neighbor if we're
		 * filtering by neighbor.
		 */
		if (rflag && recv.sin_addr.s_addr != nbraddr)
				continue;
		rptno = 0;
		gettimeofday(&t, NULL);
		/*
		 * Make sure it's DVMRP.
		 */
		if (igmp->igmp_type != 0x13)
			continue;
		/*
		 * Make sure it's a DVMRP Route Report.
		 */
		if (igmp->igmp_code != 0x2)
			continue;
		e = buf + pktlen;
		p = (unsigned char *)(igmp + 1);
		pktno++;
#ifndef SYSLOG
		/*
		 * Set up the time info for the log message.
		 */
		if (tflag) {
			sprintf(logbuf, "%d.%06d ", t.tv_sec, t.tv_usec);
		} else {
			tt = gmtime(&t.tv_sec);
			sprintf(logbuf, "%02d/%02d/%02d %02d:%02d:%02d.%03d ",
				tt->tm_mon, tt->tm_mday, tt->tm_year,
				tt->tm_hour, tt->tm_min, tt->tm_sec,
				t.tv_usec / 1000);
		}
		lbp = logbuf + strlen(logbuf);
#endif
		couldsave = 0;
		mask = 0;
		/*
		 * Parse the route report message.  A route report
		 * message contains a series of
		 * 3 bytes: netmask
		 * N bytes: network
		 * 1 byte: metric
		 * If the high bit of the metric is set, it indicates
		 * that a new netmask follows.
		 * The first byte of the netmask is assumed to be
		 * 0xff and is not transmitted.
		 * The bytes of the network address that correspond
		 * to zero bytes in the netmask are not transmitted.
		 * So if the netmask is 0xffff0000, only two bytes
		 * of network are transmitted; if the netmask is
		 * 0xfffffff0 then 4 bytes are.  See the variable
		 * "masklen".
		 */
		while (p < e) {
			oldmask = mask;
			mask = 0xff000000;
			mask |= (*p++) << 16;
			mask |= (*p++) << 8;
			mask |= (*p++);
			if (mask & 0xff)
				masklen = 4;
			else if (mask & 0xff00)
				masklen = 3;
			else if (mask & 0xff0000)
				masklen = 2;
			else
				masklen = 1;
			if (oldmask == mask &&
				    p != (unsigned char *)(igmp + 1) + 3)
				couldsave += 3;
			/*
			 * Parse all of the route entries corresponding
			 * to this mask.
			 */
			do {
				rptno++;
				addr = 0;
				for (i = 0; i < masklen; i++) {
					addr <<= 8;
					addr |= *p++;
				}
				for (; i < 4; i++) {
					addr <<= 8;
				}
				if (addr == 0 && mask == 0xff000000)
					mask = 0;
				metric = *p++;
				/*
				 * Now we have decoded the route; see
				 * if it's one of the ones we're
				 * interested in.
				 */
				if ((exactmatch && (addr == wantaddr && mask == wantmask)) ||
				    (!exactmatch && ((addr & wantmask) == wantaddr) && (mask >= wantmask))) {
					char abuf[20];
					int len = 32;
					unsigned int tmask = mask;

					/*
					 * Yup, it's interesting.
					 * Make it human-readable.
					 */

					while (((tmask & 1) == 0) && len--)
						tmask >>= 1;
					sprintf(abuf, formats[masklen - 1],
						(addr >> 24) & 0xff,
						(addr >> 16) & 0xff,
						(addr >> 8) & 0xff,
						addr & 0xff);
					sprintf(lbp,
						"%s %d %d %s/%d %s%d %d%s",
						inet_ntoa(recv.sin_addr),
						pktno, rptno,
						abuf, len,
						(metric & 0x7f) > 32 ? "PR-":"",
						(metric & 0x7f) > 32 ?
							(metric & 0x7f) - 32 :
							(metric & 0x7f),
						metric & 0x7f,
#ifdef SYSLOG
						""
#else
						"\n"
#endif
						);
					write(u, logbuf, strlen(logbuf));
				}
			} while (!(metric & 0x80));
		}
		if (cflag) {
		    sprintf(lbp,
			"%s %d %d total (could save %d bytes of %d (%d%%))%s",
			inet_ntoa(recv.sin_addr),
			pktno, rptno, couldsave, pktlen,
			couldsave * 100 / pktlen,
#ifdef SYSLOG
			""
#else
			"\n"
#endif
			);
		    write(u, logbuf, strlen(logbuf));
		}
	}
}
